

#ifndef NULL
  #define  NULL    ((void *)0)
#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Global `Adlib' operations.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


void AdlibInit(int&, char** (&)) ;
void AdlibFinalize() ;

void AdlibError(char*) ;

void AdlibPrintf(char* message, ...) ;

class Procs ;
class Dimension ;
class Coord ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `SubProcs': Process sub-arrays.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class SubProcs {
  friend class Procs ;
protected :
  //Procs* procs ;
  int rank ;

  SubProcs() ;
  SubProcs(Procs*) ;
  SubProcs(const SubProcs&) ;
public :
  int rnk() const {return rank ; }
  Procs* procs ;  // debug
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ActiveProcessGroup': the process group executing this     //
//                                control thread.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ProcTree ;
class ProcessGroup ;

class ActiveProcessGroup : public SubProcs {
  friend class Procs ;
    // Accesses `ActiveProcessGroup()',
    //          `ActiveProcessGroup(const ActiveProcessGroup&)'
  friend class ProcessGroup ;
  friend void AdlibInit(int&, char** (&)) ;
    // Accesses `ActiveProcessGroup(Procs*)'

  ActiveProcessGroup(const ActiveProcessGroup&) ;
  ActiveProcessGroup(Procs*) ;

  void deactivate(Dimension*) ;
  int deactivate(Dimension*, const int) ;
  void reactivate(Dimension*) ;

  void decrement(int*) ;
public :
  ActiveProcessGroup() ;

  int adjustProcs(Procs*) ;
  void restoreProcs() ;

  //int deactivate(Dimension*, const int) ;
  //void reactivate(Dimension*) ;

  int deactivate(Coord*) ;
  void reactivate(Coord*) ;

  int on(ProcessGroup*) ;
  void no(ProcessGroup*) ;

  int size() const ;
  int locIndex() const ;
  int procId(const int) const ;

  void spanTree(ProcTree*) ;
  void radialTree(ProcTree* tree, ProcessGroup* spine) ;
} ;


extern ActiveProcessGroup* apg ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ProcessGroup'.                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class ProcessGroup : public SubProcs {
  friend class ActiveProcessGroup ;
  friend class Cludge ;  // The name says it all...

  int* active ;
  int* coords ;
  int lead ;

  int lastAxis ;
  int testFlag ;

  int* atMask ;

  void deactivate(const Dimension*, const int) ;
  void reactivate(const Dimension*) ;

  void decrement(int*) ;
public :
  //ProcessGroup() ;
  ProcessGroup(ProcessGroup&) ;
  ProcessGroup(Procs&) ;
  ProcessGroup(ActiveProcessGroup&) ;
  ~ProcessGroup() ;

  //void deactivate(const Dimension*, const int) ;
  //void reactivate(const Dimension*) ;

  void deactivate(const Coord*) ;
  void reactivate(const Coord*) ;

  int isActive(const Dimension*) const ;

  int   procId()  const ;

  void enumerate() ;
  int  test() ;
  void next() ;

  void activeSection() ;

  int on() ;
  void no() ;

  ProcessGroup operator/(Coord&) const ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Procs {
  friend class ActiveProcessGroup ;  // accesses `physIds', `parent', `outsider'
  friend class ProcessGroup ;        // accesses `physIds'
  //friend void AdlibInit(int&, char** (&)) ; // Accesses `Procs()'
  //friend class Range ;
  //friend class Embed ;

  ActiveProcessGroup parent ;
  int rank ;
  Dimension** dims ;
  int* physIds ;
  int outsider ;

  int onStk ;

public :
  Procs() ;  // Only used in initialisation
  Procs(const int, int*) ;
  ~Procs() ;

  int rnk() const ;
  Dimension* dim(const int) const ;

  int loc() const {return !outsider ; }

  int on() ;
  void no() ;

  ProcessGroup operator/(Coord&) const ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Dimension': the process array dimension.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Dimension {
  friend class Procs ;               // accesses `Dimension(...)'
  friend class ActiveProcessGroup ;  // accesses `active', `extent', `stride', `coord'
  friend class ProcessGroup ;        // accesses `active', `extent', `stride', `axis'
  friend class Range ;               // access `extent', `blockSize', `coord'

  int extent ;
  int coord  ;
  int stride ;

  int virExtent  ;
  int blockSize  ;  // in virtual to physical mapping

  Procs* procs  ;
  int axis ;

  int active ;

  Dimension(const int, const int, const int, const int, Procs*, const int) ;
public :
  int crd() const {return coord ; }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Coord': location in a process array dimension.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Coord {
  friend class ActiveProcessGroup ;
  friend class ProcessGroup ;
protected :
  Dimension* dim ;
  int coord ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ProcsD': safer interface classes to `Procs'.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Procs0 : public Procs {
public :
  Procs0() : Procs(0, NULL) {}
} ;


class Procs1 : public Procs {
public :
  Procs1(int n1) : Procs(1, seq(n1)) {}
} ;


class Procs2 : public Procs {
public :
  Procs2(int n1, int n2) : Procs(2, seq(n1, n2)) {}
} ;


class Procs3 : public Procs {
public :
  Procs3(int n1, int n2, int n3) : Procs(3, seq(n1, n2, n3)) {}
} ;

// ---------------------------------
// Things to do
//
//   Provide a complete set of operators =, /, *, /=, *= on `ProcessGroup',
//       replacing `deactivate'/`reactivate'.
//
//   Revert to a stack based implementation of `on', to support nested
//       operations.
//
//   Change argument of 
//     ActiveProcessGroup :: ActiveProcessGroup(Procs*) ;
//   to `Procs&'?  Makes it look like a promotion.

